@@ -192,7 +192,7 @@ class FakeChange(object):
         self.addPatchset()
         self.data['submitRecords'] = self.getSubmitRecords()
 
-    def addPatchset(self, files=None, large=False):
+    def addPatchset(self, files=[], large=False):
         self.latest_patchset += 1
         if files:
             fn = files[0]
@@ -202,12 +202,15 @@ class FakeChange(object):
         c = add_fake_change_to_repo(self.project, self.branch,
                                     self.number, self.latest_patchset,
                                     msg, fn, large)
+        ps_files = [{'file': '/COMMIT_MSG',
+                     'type': 'ADDED'},
+                    {'file': 'README',
+                     'type': 'MODIFIED'}]
+        for f in files:
+            ps_files.append({'file': f, 'type': 'ADDED'})
         d = {'approvals': [],
              'createdOn': time.time(),
-             'files': [{'file': '/COMMIT_MSG',
-                        'type': 'ADDED'},
-                       {'file': 'README',
-                        'type': 'MODIFIED'}],
+             'files': ps_files,
              'number': str(self.latest_patchset),
              'ref': 'refs/changes/1/%s/%s' % (self.number,
                                               self.latest_patchset),
@@ -457,8 +460,13 @@ class FakeJenkinsJob(threading.Thread):
         if self.aborted:
             result = 'ABORTED'
 
+        changes = None
+        if 'ZUUL_CHANGE_IDS' in self.parameters:
+            changes = self.parameters['ZUUL_CHANGE_IDS']
+
         self.jenkins.fakeAddHistory(name=self.name, number=self.number,
-                                    result=result)
+                                    result=result, changes=changes)
+        self.jenkins.lock.acquire()
         self.callback.jenkins_endpoint(FakeJenkinsEvent(self.name,
                                                         self.number,
                                                         self.parameters,
@@ -470,6 +478,7 @@ class FakeJenkinsJob(threading.Thread):
                                                         'FINISHED',
                                                         result))
         self.jenkins.all_jobs.remove(self)
+        self.jenkins.lock.release()
 
 
 class FakeJenkins(object):
@@ -485,6 +494,7 @@ class FakeJenkins(object):
         self.hold_jobs_in_build = False
         self.fail_tests = {}
         self.nonexistent_jobs = []
+        self.lock = threading.Lock()
 
     def fakeEnqueue(self, job):
         self.queue.append(job)
@@ -492,8 +502,16 @@ class FakeJenkins(object):
     def fakeDequeue(self, job):
         self.queue.remove(job)
 
+    class FakeJobHistory(object):
+        def __init__(self, **kw):
+            self.__dict__.update(kw)
+
+        def __repr__(self):
+            return ("<Completed job, result: %s name: %s #%s changes: %s>" %
+                    (self.result, self.name, self.number, self.changes))
+
     def fakeAddHistory(self, **kw):
-        self.job_history.append(kw)
+        self.job_history.append(self.FakeJobHistory(**kw))
 
     def fakeRelease(self, regex=None):
         all_jobs = self.all_jobs[:]
@@ -560,7 +578,9 @@ class FakeJenkins(object):
 
     def get_queue_info(self):
         items = []
-        for job in self.queue:
+        for job in self.queue[:]:
+            self.log.debug("Queue info: %s %s" % (job.name,
+                                                  job.parameters['UUID']))
             paramstr = ''
             paramlst = []
             d = {'actions': [{'parameters': paramlst},
@@ -691,20 +711,26 @@ class testScheduler(unittest.TestCase):
                 print self.sched.result_event_queue.empty(),
                 print self.fake_gerrit.event_queue.empty(),
                 raise Exception("Timeout waiting for Zuul to settle")
+            # Make sure our fake jenkins doesn't end any jobs
+            # (and therefore, emit events) while we're checking
+            self.fake_jenkins.lock.acquire()
+            # Join ensures that the queue is empty _and_ events have been
+            # processed
             self.fake_gerrit.event_queue.join()
-            self.sched.queue_lock.acquire()
+            self.sched.trigger_event_queue.join()
+            self.sched.result_event_queue.join()
             if (self.sched.trigger_event_queue.empty() and
                 self.sched.result_event_queue.empty() and
                 self.fake_gerrit.event_queue.empty() and
                 self.fake_jenkins.fakeAllWaiting()):
-                self.sched.queue_lock.release()
+                self.fake_jenkins.lock.release()
                 self.log.debug("...settled.")
                 return
-            self.sched.queue_lock.release()
+            self.fake_jenkins.lock.release()
             self.sched.wake_event.wait(0.1)
 
     def countJobResults(self, jobs, result):
-        jobs = filter(lambda x: x['result'] == result, jobs)
+        jobs = filter(lambda x: x.result == result, jobs)
         return len(jobs)
 
     def assertEmptyQueues(self):
@@ -712,7 +738,8 @@ class testScheduler(unittest.TestCase):
         for pipeline in self.sched.pipelines.values():
             for queue in pipeline.queues:
                 if len(queue.queue) != 0:
-                    print 'queue', queue.queue
+                    print 'pipeline %s queue %s contents %s' % (
+                        pipeline.name, queue.name, queue.queue)
                 assert len(queue.queue) == 0
                 if len(queue.severed_heads) != 0:
                     print 'heads', queue.severed_heads
@@ -725,13 +752,13 @@ class testScheduler(unittest.TestCase):
         self.fake_gerrit.addEvent(A.addApproval('APRV', 1))
         self.waitUntilSettled()
         jobs = self.fake_jenkins.job_history
-        job_names = [x['name'] for x in jobs]
+        job_names = [x.name for x in jobs]
         assert 'project-merge' in job_names
         assert 'project-test1' in job_names
         assert 'project-test2' in job_names
-        assert jobs[0]['result'] == 'SUCCESS'
-        assert jobs[1]['result'] == 'SUCCESS'
-        assert jobs[2]['result'] == 'SUCCESS'
+        assert jobs[0].result == 'SUCCESS'
+        assert jobs[1].result == 'SUCCESS'
+        assert jobs[2].result == 'SUCCESS'
         assert A.data['status'] == 'MERGED'
         assert A.reported == 2
         self.assertEmptyQueues()
@@ -935,6 +962,11 @@ class testScheduler(unittest.TestCase):
         self.fake_jenkins.fakeRelease()
         self.waitUntilSettled()
 
+        for x in jobs:
+            print x
+        for x in finished_jobs:
+            print x
+
         assert len(jobs) == 0
         assert len(finished_jobs) == 15
         assert A.data['status'] == 'NEW'
@@ -1162,7 +1194,7 @@ class testScheduler(unittest.TestCase):
         self.waitUntilSettled()
 
         jobs = self.fake_jenkins.job_history
-        job_names = [x['name'] for x in jobs]
+        job_names = [x.name for x in jobs]
         assert len(jobs) == 1
         assert 'project-post' in job_names
         self.assertEmptyQueues()
@@ -1322,7 +1354,7 @@ class testScheduler(unittest.TestCase):
         self.assertEmptyQueues()
 
     def test_head_is_dequeued_once(self):
-        "Test that if a change at the head fails it is dequeud only once"
+        "Test that if a change at the head fails it is dequeued only once"
         # If it's dequeued more than once, we should see extra
         # aborted jobs.
         self.fake_jenkins.hold_jobs_in_build = True
@@ -1403,9 +1435,9 @@ class testScheduler(unittest.TestCase):
 
         assert A.data['status'] == 'MERGED'
         assert A.reported == 2
-        assert finished_jobs[0]['result'] == 'SUCCESS'
-        assert finished_jobs[1]['result'] == 'SUCCESS'
-        assert finished_jobs[2]['result'] == 'FAILURE'
+        assert finished_jobs[0].result == 'SUCCESS'
+        assert finished_jobs[1].result == 'SUCCESS'
+        assert finished_jobs[2].result == 'FAILURE'
         self.assertEmptyQueues()
 
     def test_check_queue_success(self):
@@ -1419,9 +1451,9 @@ class testScheduler(unittest.TestCase):
 
         assert A.data['status'] == 'NEW'
         assert A.reported == 1
-        assert finished_jobs[0]['result'] == 'SUCCESS'
-        assert finished_jobs[1]['result'] == 'SUCCESS'
-        assert finished_jobs[2]['result'] == 'SUCCESS'
+        assert finished_jobs[0].result == 'SUCCESS'
+        assert finished_jobs[1].result == 'SUCCESS'
+        assert finished_jobs[2].result == 'SUCCESS'
         self.assertEmptyQueues()
 
     def test_check_queue_failure(self):
@@ -1436,9 +1468,9 @@ class testScheduler(unittest.TestCase):
 
         assert A.data['status'] == 'NEW'
         assert A.reported == 1
-        assert finished_jobs[0]['result'] == 'SUCCESS'
-        assert finished_jobs[1]['result'] == 'SUCCESS'
-        assert finished_jobs[2]['result'] == 'FAILURE'
+        assert finished_jobs[0].result == 'SUCCESS'
+        assert finished_jobs[1].result == 'SUCCESS'
+        assert finished_jobs[2].result == 'FAILURE'
         self.assertEmptyQueues()
 
     def test_dependent_behind_dequeue(self):
@@ -1545,13 +1577,13 @@ class testScheduler(unittest.TestCase):
         self.fake_gerrit.addEvent(A.addApproval('APRV', 1))
         self.waitUntilSettled()
         jobs = self.fake_jenkins.job_history
-        job_names = [x['name'] for x in jobs]
+        job_names = [x.name for x in jobs]
         assert 'project-merge' in job_names
         assert 'project-test1' in job_names
         assert 'project-test2' in job_names
-        assert jobs[0]['result'] == 'SUCCESS'
-        assert jobs[1]['result'] == 'SUCCESS'
-        assert jobs[2]['result'] == 'SUCCESS'
+        assert jobs[0].result == 'SUCCESS'
+        assert jobs[1].result == 'SUCCESS'
+        assert jobs[2].result == 'SUCCESS'
         assert A.data['status'] == 'MERGED'
         assert A.reported == 2
         self.assertEmptyQueues()
@@ -1564,13 +1596,13 @@ class testScheduler(unittest.TestCase):
         self.fake_gerrit.addEvent(A.addApproval('APRV', 1))
         self.waitUntilSettled()
         jobs = self.fake_jenkins.job_history
-        job_names = [x['name'] for x in jobs]
+        job_names = [x.name for x in jobs]
         assert 'project-merge' in job_names
         assert 'project-test1' in job_names
         assert 'project-test2' in job_names
-        assert jobs[0]['result'] == 'SUCCESS'
-        assert jobs[1]['result'] == 'SUCCESS'
-        assert jobs[2]['result'] == 'SUCCESS'
+        assert jobs[0].result == 'SUCCESS'
+        assert jobs[1].result == 'SUCCESS'
+        assert jobs[2].result == 'SUCCESS'
         assert A.data['status'] == 'MERGED'
         assert A.reported == 2
         self.assertEmptyQueues()
@@ -1589,13 +1621,13 @@ class testScheduler(unittest.TestCase):
         self.fake_gerrit.addEvent(A.addApproval('APRV', 1))
         self.waitUntilSettled()
         jobs = self.fake_jenkins.job_history
-        job_names = [x['name'] for x in jobs]
+        job_names = [x.name for x in jobs]
         assert 'project1-merge' in job_names
         assert 'project1-test1' in job_names
         assert 'project1-test2' in job_names
-        assert jobs[0]['result'] == 'SUCCESS'
-        assert jobs[1]['result'] == 'SUCCESS'
-        assert jobs[2]['result'] == 'SUCCESS'
+        assert jobs[0].result == 'SUCCESS'
+        assert jobs[1].result == 'SUCCESS'
+        assert jobs[2].result == 'SUCCESS'
         assert A.data['status'] == 'MERGED'
         assert A.reported == 2
         self.assertEmptyQueues()
@@ -1612,7 +1644,7 @@ class testScheduler(unittest.TestCase):
         while A.reported < 2:
             self.waitUntilSettled()
         jobs = self.fake_jenkins.job_history
-        job_names = [x['name'] for x in jobs]
+        job_names = [x.name for x in jobs]
         assert not job_names
         assert A.data['status'] == 'NEW'
         assert A.reported == 2
@@ -1625,13 +1657,340 @@ class testScheduler(unittest.TestCase):
         self.fake_gerrit.addEvent(A.addApproval('APRV', 1))
         self.waitUntilSettled()
         jobs = self.fake_jenkins.job_history
-        job_names = [x['name'] for x in jobs]
+        job_names = [x.name for x in jobs]
         assert 'project-merge' in job_names
         assert 'project-test1' in job_names
         assert 'project-test2' in job_names
-        assert jobs[0]['result'] == 'SUCCESS'
-        assert jobs[1]['result'] == 'SUCCESS'
-        assert jobs[2]['result'] == 'SUCCESS'
+        assert jobs[0].result == 'SUCCESS'
+        assert jobs[1].result == 'SUCCESS'
+        assert jobs[2].result == 'SUCCESS'
         assert A.data['status'] == 'MERGED'
         assert A.reported == 2
         self.assertEmptyQueues()
+
+    def test_single_nonexistent_post_job(self):
+        "Test launching a single post job that doesn't exist"
+        self.fake_jenkins.nonexistent_jobs.append('project-post')
+        self.jenkins.launch_retry_timeout = 0.1
+
+        e = {
+            "type": "ref-updated",
+            "submitter": {
+                "name": "User Name",
+            },
+            "refUpdate": {
+                "oldRev": "90f173846e3af9154517b88543ffbd1691f31366",
+                "newRev": "d479a0bfcb34da57a31adb2a595c0cf687812543",
+                "refName": "master",
+                "project": "org/project",
+            }
+        }
+        self.fake_gerrit.addEvent(e)
+        self.waitUntilSettled()
+
+        jobs = self.fake_jenkins.job_history
+        assert len(jobs) == 0
+        self.assertEmptyQueues()
+
+    def test_new_patchset_dequeues_old(self):
+        "Test that a new patchset causes the old to be dequeued"
+        # D -> C (depends on B) -> B (depends on A) -> A -> M
+        self.fake_jenkins.hold_jobs_in_build = True
+
+        M = self.fake_gerrit.addFakeChange('org/project', 'master', 'M')
+        M.setMerged()
+
+        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
+        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
+        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
+        D = self.fake_gerrit.addFakeChange('org/project', 'master', 'D')
+        A.addApproval('CRVW', 2)
+        B.addApproval('CRVW', 2)
+        C.addApproval('CRVW', 2)
+        D.addApproval('CRVW', 2)
+
+        C.setDependsOn(B, 1)
+        B.setDependsOn(A, 1)
+        A.setDependsOn(M, 1)
+
+        self.fake_gerrit.addEvent(A.addApproval('APRV', 1))
+        self.fake_gerrit.addEvent(B.addApproval('APRV', 1))
+        self.fake_gerrit.addEvent(C.addApproval('APRV', 1))
+        self.fake_gerrit.addEvent(D.addApproval('APRV', 1))
+        self.waitUntilSettled()
+
+        B.addPatchset()
+        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(2))
+        self.waitUntilSettled()
+
+        self.fake_jenkins.hold_jobs_in_build = False
+        self.fake_jenkins.fakeRelease()
+        self.waitUntilSettled()
+
+        jobs = self.fake_jenkins.all_jobs
+        finished_jobs = self.fake_jenkins.job_history
+
+        for x in jobs:
+            print x
+        for x in finished_jobs:
+            print x
+
+        assert A.data['status'] == 'MERGED'
+        assert A.reported == 2
+        assert B.data['status'] == 'NEW'
+        assert B.reported == 2
+        assert C.data['status'] == 'NEW'
+        assert C.reported == 2
+        assert D.data['status'] == 'MERGED'
+        assert D.reported == 2
+        assert len(finished_jobs) == 9  # 3 each for A, B, D.
+        self.assertEmptyQueues()
+
+    def test_new_patchset_dequeues_old_on_head(self):
+        "Test that a new patchset causes the old to be dequeued (at head)"
+        # D -> C (depends on B) -> B (depends on A) -> A -> M
+        self.fake_jenkins.hold_jobs_in_build = True
+
+        M = self.fake_gerrit.addFakeChange('org/project', 'master', 'M')
+        M.setMerged()
+        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
+        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
+        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
+        D = self.fake_gerrit.addFakeChange('org/project', 'master', 'D')
+        A.addApproval('CRVW', 2)
+        B.addApproval('CRVW', 2)
+        C.addApproval('CRVW', 2)
+        D.addApproval('CRVW', 2)
+
+        C.setDependsOn(B, 1)
+        B.setDependsOn(A, 1)
+        A.setDependsOn(M, 1)
+
+        self.fake_gerrit.addEvent(A.addApproval('APRV', 1))
+        self.fake_gerrit.addEvent(B.addApproval('APRV', 1))
+        self.fake_gerrit.addEvent(C.addApproval('APRV', 1))
+        self.fake_gerrit.addEvent(D.addApproval('APRV', 1))
+        self.waitUntilSettled()
+
+        A.addPatchset()
+        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(2))
+        self.waitUntilSettled()
+
+        self.fake_jenkins.hold_jobs_in_build = False
+        self.fake_jenkins.fakeRelease()
+        self.waitUntilSettled()
+
+        jobs = self.fake_jenkins.all_jobs
+        finished_jobs = self.fake_jenkins.job_history
+
+        for x in jobs:
+            print x
+        for x in finished_jobs:
+            print x
+
+        assert A.data['status'] == 'NEW'
+        assert A.reported == 2
+        assert B.data['status'] == 'NEW'
+        assert B.reported == 2
+        assert C.data['status'] == 'NEW'
+        assert C.reported == 2
+        assert D.data['status'] == 'MERGED'
+        assert D.reported == 2
+        assert len(finished_jobs) == 7
+        self.assertEmptyQueues()
+
+    def test_new_patchset_dequeues_old_without_dependents(self):
+        "Test that a new patchset causes only the old to be dequeued"
+        self.fake_jenkins.hold_jobs_in_build = True
+
+        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
+        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
+        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
+        A.addApproval('CRVW', 2)
+        B.addApproval('CRVW', 2)
+        C.addApproval('CRVW', 2)
+
+        self.fake_gerrit.addEvent(C.addApproval('APRV', 1))
+        self.fake_gerrit.addEvent(B.addApproval('APRV', 1))
+        self.fake_gerrit.addEvent(A.addApproval('APRV', 1))
+        self.waitUntilSettled()
+
+        B.addPatchset()
+        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(2))
+        self.waitUntilSettled()
+
+        self.fake_jenkins.hold_jobs_in_build = False
+        self.fake_jenkins.fakeRelease()
+        self.waitUntilSettled()
+
+        jobs = self.fake_jenkins.all_jobs
+        finished_jobs = self.fake_jenkins.job_history
+
+        for x in jobs:
+            print x
+        for x in finished_jobs:
+            print x
+
+        assert A.data['status'] == 'MERGED'
+        assert A.reported == 2
+        assert B.data['status'] == 'NEW'
+        assert B.reported == 2
+        assert C.data['status'] == 'MERGED'
+        assert C.reported == 2
+        assert len(finished_jobs) == 9
+        self.assertEmptyQueues()
+
+    def test_new_patchset_dequeues_old_independent_queue(self):
+        "Test that a new patchset causes the old to be dequeued (independent)"
+        self.fake_jenkins.hold_jobs_in_build = True
+
+        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
+        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
+        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
+        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
+        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
+        self.fake_gerrit.addEvent(C.getPatchsetCreatedEvent(1))
+        self.waitUntilSettled()
+
+        B.addPatchset()
+        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(2))
+        self.waitUntilSettled()
+
+        self.fake_jenkins.hold_jobs_in_build = False
+        self.fake_jenkins.fakeRelease()
+        self.waitUntilSettled()
+
+        jobs = self.fake_jenkins.all_jobs
+        finished_jobs = self.fake_jenkins.job_history
+
+        for x in jobs:
+            print x
+        for x in finished_jobs:
+            print x
+
+        assert A.data['status'] == 'NEW'
+        assert A.reported == 1
+        assert B.data['status'] == 'NEW'
+        assert B.reported == 1
+        assert C.data['status'] == 'NEW'
+        assert C.reported == 1
+        assert len(finished_jobs) == 10
+        assert self.countJobResults(finished_jobs, 'ABORTED') == 1
+        self.assertEmptyQueues()
+
+    def test_zuul_refs(self):
+        "Test that zuul refs exist and have the right changes"
+        self.fake_jenkins.hold_jobs_in_build = True
+
+        M1 = self.fake_gerrit.addFakeChange('org/project1', 'master', 'M1')
+        M1.setMerged()
+        M2 = self.fake_gerrit.addFakeChange('org/project2', 'master', 'M2')
+        M2.setMerged()
+
+        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
+        B = self.fake_gerrit.addFakeChange('org/project1', 'master', 'B')
+        C = self.fake_gerrit.addFakeChange('org/project2', 'master', 'C')
+        D = self.fake_gerrit.addFakeChange('org/project2', 'master', 'D')
+        A.addApproval('CRVW', 2)
+        B.addApproval('CRVW', 2)
+        C.addApproval('CRVW', 2)
+        D.addApproval('CRVW', 2)
+        self.fake_gerrit.addEvent(A.addApproval('APRV', 1))
+        self.fake_gerrit.addEvent(B.addApproval('APRV', 1))
+        self.fake_gerrit.addEvent(C.addApproval('APRV', 1))
+        self.fake_gerrit.addEvent(D.addApproval('APRV', 1))
+
+        self.waitUntilSettled()
+        self.fake_jenkins.fakeRelease('.*-merge')
+        self.waitUntilSettled()
+        self.fake_jenkins.fakeRelease('.*-merge')
+        self.waitUntilSettled()
+        self.fake_jenkins.fakeRelease('.*-merge')
+        self.waitUntilSettled()
+        self.fake_jenkins.fakeRelease('.*-merge')
+        self.waitUntilSettled()
+
+        jobs = self.fake_jenkins.all_jobs
+        finished_jobs = self.fake_jenkins.job_history
+
+        a_zref = b_zref = c_zref = d_zref = None
+        for x in jobs:
+            if x.parameters['ZUUL_CHANGE'] == '3':
+                a_zref = x.parameters['ZUUL_REF']
+            if x.parameters['ZUUL_CHANGE'] == '4':
+                b_zref = x.parameters['ZUUL_REF']
+            if x.parameters['ZUUL_CHANGE'] == '5':
+                c_zref = x.parameters['ZUUL_REF']
+            if x.parameters['ZUUL_CHANGE'] == '6':
+                d_zref = x.parameters['ZUUL_REF']
+
+        # There are... four... refs.
+        assert a_zref is not None
+        assert b_zref is not None
+        assert c_zref is not None
+        assert d_zref is not None
+
+        # And they should all be different
+        refs = set([a_zref, b_zref, c_zref, d_zref])
+        assert len(refs) == 4
+
+        # a ref should have a, not b, and should not be in project2
+        assert ref_has_change(a_zref, A)
+        assert not ref_has_change(a_zref, B)
+        assert not ref_has_change(a_zref, M2)
+
+        # b ref should have a and b, and should not be in project2
+        assert ref_has_change(b_zref, A)
+        assert ref_has_change(b_zref, B)
+        assert not ref_has_change(b_zref, M2)
+
+        # c ref should have a and b in 1, c in 2
+        assert ref_has_change(c_zref, A)
+        assert ref_has_change(c_zref, B)
+        assert ref_has_change(c_zref, C)
+        assert not ref_has_change(c_zref, D)
+
+        # d ref should have a and b in 1, c and d in 2
+        assert ref_has_change(d_zref, A)
+        assert ref_has_change(d_zref, B)
+        assert ref_has_change(d_zref, C)
+        assert ref_has_change(d_zref, D)
+
+        self.fake_jenkins.hold_jobs_in_build = False
+        self.fake_jenkins.fakeRelease()
+        self.waitUntilSettled()
+
+        assert A.data['status'] == 'MERGED'
+        assert A.reported == 2
+        assert B.data['status'] == 'MERGED'
+        assert B.reported == 2
+        assert C.data['status'] == 'MERGED'
+        assert C.reported == 2
+        assert D.data['status'] == 'MERGED'
+        assert D.reported == 2
+        self.assertEmptyQueues()
+
+    def test_file_jobs(self):
+        "Test that file jobs run only when appropriate"
+        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
+        A.addPatchset(['pip-requires'])
+        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
+        A.addApproval('CRVW', 2)
+        B.addApproval('CRVW', 2)
+        self.fake_gerrit.addEvent(A.addApproval('APRV', 1))
+        self.fake_gerrit.addEvent(B.addApproval('APRV', 1))
+        self.waitUntilSettled()
+
+        jobs = self.fake_jenkins.all_jobs
+        finished_jobs = self.fake_jenkins.job_history
+
+        testfile_jobs = [x for x in finished_jobs
+                         if x.name == 'project-testfile']
+
+        assert len(testfile_jobs) == 1
+        assert testfile_jobs[0].changes == '1,2'
+        assert A.data['status'] == 'MERGED'
+        assert A.reported == 2
+        assert B.data['status'] == 'MERGED'
+        assert B.reported == 2
+        self.assertEmptyQueues()
