@@ -49,9 +49,9 @@ class JenkinsCallback(threading.Thread):
 
     def app(self, environ, start_response):
         request = Request(environ)
-        start_response('200 OK', [('content-type', 'text/html')])
         if request.path == '/jenkins_endpoint':
             self.jenkins_endpoint(request)
+            start_response('200 OK', [('content-type', 'text/html')])
             return ['Zuul good.']
         elif request.path == '/status':
             try:
@@ -59,8 +59,19 @@ class JenkinsCallback(threading.Thread):
             except:
                 self.log.exception("Exception formatting status:")
                 raise
+            start_response('200 OK', [('content-type', 'text/html')])
+            return [ret]
+        elif request.path == '/status.json':
+            try:
+                ret = self.jenkins.sched.formatStatusJSON()
+            except:
+                self.log.exception("Exception formatting status:")
+                raise
+            start_response('200 OK', [('content-type', 'application/json'),
+                                      ('Access-Control-Allow-Origin', '*')])
             return [ret]
         else:
+            start_response('200 OK', [('content-type', 'text/html')])
             return ['Zuul good.']
 
     def jenkins_endpoint(self, request):
@@ -122,6 +133,7 @@ STOP_BUILD = 'job/%(name)s/%(number)s/stop'
 CANCEL_QUEUE = 'queue/item/%(number)s/cancelQueue'
 BUILD_INFO = 'job/%(name)s/%(number)s/api/json?depth=0'
 BUILD_DESCRIPTION = 'job/%(name)s/%(number)s/submitDescription'
+DEBUG = False
 
 
 class ExtendedJenkins(jenkins.Jenkins):
@@ -134,8 +146,9 @@ class ExtendedJenkins(jenkins.Jenkins):
                 req.add_header('Authorization', self.auth)
             return urllib2.urlopen(req).read()
         except urllib2.HTTPError, e:
-            print e.msg
-            print e.fp.read()
+            if DEBUG:
+                print e.msg
+                print e.fp.read()
             raise
 
     def stop_build(self, name, number):
@@ -160,6 +173,7 @@ class ExtendedJenkins(jenkins.Jenkins):
         # Jenkins returns a 302 from this URL, unless Referer is not set,
         # then you get a 404.
         request = urllib2.Request(self.server + CANCEL_QUEUE % locals(),
+                                  urllib.urlencode({}),
                                   headers={'Referer': self.server})
         self.jenkins_open(request)
 
@@ -313,16 +327,9 @@ class Jenkins(object):
                                "declaring lost" % build)
                 # To keep the queue moving, declare this as a lost build
                 # so that the change will get dropped.
-                t = threading.Thread(target=self.declareBuildLost,
-                                     args=(build,))
-                t.start()
+                self.onBuildCompleted(build.uuid, 'LOST', None, None)
         return build
 
-    def declareBuildLost(self, build):
-        # Call this from a new thread to invoke onBuildCompleted from
-        # a thread that has the queue lock.
-        self.onBuildCompleted(build.uuid, 'LOST', None, None)
-
     def findBuildInQueue(self, build):
         for item in self.jenkins.get_queue_info():
             if 'actions' not in item:
