@@ -27,6 +27,10 @@ class Pipeline(object):
     """A top-level pipeline such as check, gate, post, etc."""
     def __init__(self, name):
         self.name = name
+        self.description = None
+        self.failure_message = None
+        self.success_message = None
+        self.dequeue_on_new_patchset = True
         self.job_trees = {}  # project -> JobTree
         self.manager = None
         self.queues = []
@@ -119,7 +123,7 @@ class Pipeline(object):
             if not job.voting:
                 continue
             build = changeish.current_build_set.getBuild(job.name)
-            if build and build.result == 'FAILURE':
+            if build and build.result and (build.result != 'SUCCESS'):
                 return True
         return False
 
@@ -186,6 +190,24 @@ class Pipeline(object):
                 ret += self.formatStatus(head, html=True)
         return ret
 
+    def formatStatusJSON(self):
+        j_pipeline = dict(name=self.name,
+                          description=self.description)
+        j_queues = []
+        j_pipeline['change_queues'] = j_queues
+        for queue in self.queues:
+            j_queue = dict(name=queue.name)
+            j_queues.append(j_queue)
+            j_queue['heads'] = []
+            for head in queue.getHeads():
+                j_changes = []
+                c = head
+                while c:
+                    j_changes.append(self.formatChangeJSON(c))
+                    c = c.change_behind
+                j_queue['heads'].append(j_changes)
+        return j_pipeline
+
     def formatStatus(self, changeish, indent=0, html=False):
         indent_str = ' ' * indent
         ret = ''
@@ -224,6 +246,31 @@ class Pipeline(object):
             ret += self.formatStatus(changeish.change_behind, indent + 2, html)
         return ret
 
+    def formatChangeJSON(self, changeish):
+        ret = {}
+        if hasattr(changeish, 'url') and changeish.url is not None:
+            ret['url'] = changeish.url
+        else:
+            ret['url'] = None
+        ret['id'] = changeish._id()
+        ret['project'] = changeish.project.name
+        ret['jobs'] = []
+        for job in self.getJobs(changeish):
+            build = changeish.current_build_set.getBuild(job.name)
+            if build:
+                result = build.result
+                url = build.url
+            else:
+                result = None
+                url = None
+            ret['jobs'].append(
+                dict(
+                    name=job.name,
+                    url=url,
+                    result=result,
+                    voting=job.voting))
+        return ret
+
 
 class ChangeQueue(object):
     """DependentPipelines have multiple parallel queues shared by
@@ -259,6 +306,7 @@ class ChangeQueue(object):
             change.change_ahead = self.queue[-1]
             change.change_ahead.change_behind = change
         self.queue.append(change)
+        change.enqueue_time = time.time()
 
     def dequeueChange(self, change):
         if change in self.queue:
@@ -271,6 +319,7 @@ class ChangeQueue(object):
             change.change_behind.change_ahead = change.change_ahead
         change.change_ahead = None
         change.change_behind = None
+        change.dequeue_time = time.time()
 
     def addSeveredHead(self, change):
         self.severed_heads.append(change)
@@ -314,11 +363,15 @@ class Job(object):
         self.name = name
         self.failure_message = None
         self.success_message = None
+        self.failure_pattern = None
+        self.success_pattern = None
         self.parameter_function = None
         self.hold_following_changes = False
         self.voting = True
         self.branches = []
         self._branches = []
+        self.files = []
+        self._files = []
 
     def __str__(self):
         return self.name
@@ -329,21 +382,36 @@ class Job(object):
     def copy(self, other):
         self.failure_message = other.failure_message
         self.success_message = other.success_message
+        self.failure_pattern = other.failure_pattern
+        self.success_pattern = other.success_pattern
         self.parameter_function = other.parameter_function
         self.hold_following_changes = other.hold_following_changes
         self.voting = other.voting
         self.branches = other.branches[:]
         self._branches = other._branches[:]
+        self.files = other.files[:]
+        self._files = other._files[:]
 
     def changeMatches(self, change):
-        if not self.branches:
-            return True
+        matches_branch = False
         for branch in self.branches:
             if hasattr(change, 'branch') and branch.match(change.branch):
-                return True
+                matches_branch = True
             if hasattr(change, 'ref') and branch.match(change.ref):
-                return True
-        return False
+                matches_branch = True
+        if self.branches and not matches_branch:
+            return False
+
+        matches_file = False
+        for f in self.files:
+            if hasattr(change, 'files'):
+                for cf in change.files:
+                    if f.match(cf):
+                        matches_file = True
+        if self.files and not matches_file:
+            return False
+
+        return True
 
 
 class JobTree(object):
@@ -388,6 +456,8 @@ class Build(object):
         self.result = None
         self.build_set = None
         self.launch_time = time.time()
+        self.start_time = None
+        self.end_time = None
 
     def __repr__(self):
         return '<Build %s of %s>' % (self.uuid, self.job.name)
@@ -442,10 +512,15 @@ class Changeish(object):
         self.build_sets.append(self.current_build_set)
         self.change_ahead = None
         self.change_behind = None
+        self.enqueue_time = None
+        self.dequeue_time = None
 
     def equals(self, other):
         raise NotImplementedError()
 
+    def isUpdateOf(self, other):
+        raise NotImplementedError()
+
     def filterJobs(self, jobs):
         return filter(lambda job: job.changeMatches(self), jobs)
 
@@ -472,6 +547,7 @@ class Change(Changeish):
         self.patchset = None
         self.refspec = None
 
+        self.files = []
         self.reported = False
         self.needs_change = None
         self.needed_by_changes = []
@@ -490,6 +566,12 @@ class Change(Changeish):
             return True
         return False
 
+    def isUpdateOf(self, other):
+        if ((hasattr(other, 'number') and self.number == other.number) and
+            (hasattr(other, 'patchset') and self.patchset > other.patchset)):
+            return True
+        return False
+
     def setReportedResult(self, result):
         self.current_build_set.result = result
 
@@ -506,6 +588,21 @@ class Ref(Changeish):
     def _id(self):
         return self.newrev
 
+    def __repr__(self):
+        rep = None
+        if self.newrev == '0000000000000000000000000000000000000000':
+            rep = '<Ref 0x%x deletes %s from %s' % (
+                  id(self), self.ref, self.oldrev)
+        elif self.oldrev == '0000000000000000000000000000000000000000':
+            rep = '<Ref 0x%x creates %s on %s>' % (
+                  id(self), self.ref, self.newrev)
+        else:
+            # Catch all
+            rep = '<Ref 0x%x %s updated %s..%s>' % (
+                  id(self), self.ref, self.oldrev, self.newrev)
+
+        return rep
+
     def equals(self, other):
         if (self.project == other.project
             and self.ref == other.ref
@@ -513,6 +610,9 @@ class Ref(Changeish):
             return True
         return False
 
+    def isUpdateOf(self, other):
+        return False
+
 
 class TriggerEvent(object):
     def __init__(self):
@@ -520,6 +620,8 @@ class TriggerEvent(object):
         # common
         self.type = None
         self.project_name = None
+        # Representation of the user account that performed the event.
+        self.account = None
         # patchset-created, comment-added, etc.
         self.change_number = None
         self.change_url = None
@@ -559,20 +661,24 @@ class TriggerEvent(object):
             change.ref = self.ref
             change.oldrev = self.oldrev
             change.newrev = self.newrev
+            change.url = trigger.getGitwebUrl(project, sha=self.newrev)
 
         return change
 
 
 class EventFilter(object):
     def __init__(self, types=[], branches=[], refs=[], approvals={},
-                 comment_filters=[]):
+                 comment_filters=[], email_filters=[]):
         self._types = types
         self._branches = branches
         self._refs = refs
+        self._comment_filters = comment_filters
+        self._email_filters = email_filters
         self.types = [re.compile(x) for x in types]
         self.branches = [re.compile(x) for x in branches]
         self.refs = [re.compile(x) for x in refs]
         self.comment_filters = [re.compile(x) for x in comment_filters]
+        self.email_filters = [re.compile(x) for x in email_filters]
         self.approvals = approvals
 
     def __repr__(self):
@@ -587,6 +693,10 @@ class EventFilter(object):
         if self.approvals:
             ret += ' approvals: %s' % ', '.join(
                 ['%s:%s' % a for a in self.approvals.items()])
+        if self._comment_filters:
+            ret += ' comment_filters: %s' % ', '.join(self._comment_filters)
+        if self._email_filters:
+            ret += ' email_filters: %s' % ', '.join(self._email_filters)
         ret += '>'
 
         return ret
@@ -624,11 +734,24 @@ class EventFilter(object):
         matches_comment_filter = False
         for comment_filter in self.comment_filters:
             if (event.comment is not None and
-                    comment_filter.search(event.comment)):
+                comment_filter.search(event.comment)):
                 matches_comment_filter = True
         if self.comment_filters and not matches_comment_filter:
             return False
 
+        # We better have an account provided by Gerrit to do
+        # email filtering.
+        if event.account is not None:
+            account_email = event.account.get('email')
+            # email_filters are ORed
+            matches_email_filter = False
+            for email_filter in self.email_filters:
+                if (account_email is not None and
+                    email_filter.search(account_email)):
+                    matches_email_filter = True
+            if self.email_filters and not matches_email_filter:
+                return False
+
         # approvals are ANDed
         for category, value in self.approvals.items():
             matches_approval = False
