@@ -1,4 +1,5 @@
 # Copyright 2012 Hewlett-Packard Development Company, L.P.
+# Copyright 2013 OpenStack Foundation
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
@@ -12,18 +13,24 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
+import extras
+import json
 import logging
 import os
 import pickle
 import Queue
 import re
 import threading
+import time
 import yaml
 
+import layoutvalidator
 import model
 from model import Pipeline, Job, Project, ChangeQueue, EventFilter
 import merger
 
+statsd = extras.try_import('statsd.statsd')
+
 
 class Scheduler(threading.Thread):
     log = logging.getLogger("zuul.Scheduler")
@@ -32,7 +39,6 @@ class Scheduler(threading.Thread):
         threading.Thread.__init__(self)
         self.wake_event = threading.Event()
         self.reconfigure_complete_event = threading.Event()
-        self.queue_lock = threading.Lock()
         self._pause = False
         self._reconfigure = False
         self._exit = False
@@ -54,6 +60,10 @@ class Scheduler(threading.Thread):
         self._stopped = True
         self.wake_event.set()
 
+    def testConfig(self, config_path):
+        self._init()
+        self._parseConfig(config_path)
+
     def _parseConfig(self, config_path):
         def toList(item):
             if not item:
@@ -70,6 +80,9 @@ class Scheduler(threading.Thread):
         config_file = open(config_path)
         data = yaml.load(config_file)
 
+        validator = layoutvalidator.LayoutValidator()
+        validator.validate(data)
+
         self._config_env = {}
         for include in data.get('includes', []):
             if 'python-file' in include:
@@ -82,6 +95,14 @@ class Scheduler(threading.Thread):
 
         for conf_pipeline in data.get('pipelines', []):
             pipeline = Pipeline(conf_pipeline['name'])
+            pipeline.description = conf_pipeline.get('description')
+            pipeline.failure_message = conf_pipeline.get('failure-message',
+                                                         "Build failed.")
+            pipeline.success_message = conf_pipeline.get('success-message',
+                                                         "Build succeeded.")
+            pipeline.dequeue_on_new_patchset = conf_pipeline.get(
+                'dequeue-on-new-patchset',
+                True)
             manager = globals()[conf_pipeline['manager']](self, pipeline)
             pipeline.setManager(manager)
 
@@ -99,10 +120,12 @@ class Scheduler(threading.Thread):
                                 refs=toList(trigger.get('ref')),
                                 approvals=approvals,
                                 comment_filters=
-                                toList(trigger.get('comment_filter')))
+                                toList(trigger.get('comment_filter')),
+                                email_filters=
+                                toList(trigger.get('email_filter')))
                 manager.event_filters.append(f)
 
-        for config_job in data['jobs']:
+        for config_job in data.get('jobs', []):
             job = self.getJob(config_job['name'])
             # Be careful to only set attributes explicitly present on
             # this job, to avoid squashing attributes set by a meta-job.
@@ -112,6 +135,12 @@ class Scheduler(threading.Thread):
             m = config_job.get('success-message', None)
             if m:
                 job.success_message = m
+            m = config_job.get('failure-pattern', None)
+            if m:
+                job.failure_pattern = m
+            m = config_job.get('success-pattern', None)
+            if m:
+                job.success_pattern = m
             m = config_job.get('hold-following-changes', False)
             if m:
                 job.hold_following_changes = True
@@ -128,6 +157,10 @@ class Scheduler(threading.Thread):
             if branches:
                 job._branches = branches
                 job.branches = [re.compile(x) for x in branches]
+            files = toList(config_job.get('files'))
+            if files:
+                job._files = files
+                job.files = [re.compile(x) for x in files]
 
         def add_jobs(job_tree, config_jobs):
             for job in config_jobs:
@@ -141,7 +174,7 @@ class Scheduler(threading.Thread):
                 if isinstance(job, str):
                     job_tree.addJob(self.getJob(job))
 
-        for config_project in data['projects']:
+        for config_project in data.get('projects', []):
             project = Project(config_project['name'])
             self.projects[config_project['name']] = project
             mode = config_project.get('merge-mode')
@@ -157,23 +190,25 @@ class Scheduler(threading.Thread):
         # metajobs so that getJob isn't doing anything weird.
         self.metajobs = {}
 
-        # TODO(jeblair): check that we don't end up with jobs like
-        # "foo - bar" because a ':' is missing in the yaml for a dependent job
         for pipeline in self.pipelines.values():
             pipeline.manager._postConfig()
 
+    def _setupMerger(self):
         if self.config.has_option('zuul', 'git_dir'):
             merge_root = self.config.get('zuul', 'git_dir')
         else:
             merge_root = '/var/lib/zuul/git'
+
         if self.config.has_option('zuul', 'push_change_refs'):
             push_refs = self.config.getboolean('zuul', 'push_change_refs')
         else:
             push_refs = False
+
         if self.config.has_option('gerrit', 'sshkey'):
             sshkey = self.config.get('gerrit', 'sshkey')
         else:
             sshkey = None
+
         self.merger = merger.Merger(self.trigger, merge_root, push_refs,
                                     sshkey)
         for project in self.projects.values():
@@ -204,24 +239,37 @@ class Scheduler(threading.Thread):
 
     def addEvent(self, event):
         self.log.debug("Adding trigger event: %s" % event)
-        self.queue_lock.acquire()
+        try:
+            if statsd:
+                statsd.incr('gerrit.event.%s' % event.type)
+        except:
+            self.log.exception("Exception reporting event stats")
         self.trigger_event_queue.put(event)
-        self.queue_lock.release()
         self.wake_event.set()
+        self.log.debug("Done adding trigger event: %s" % event)
 
     def onBuildStarted(self, build):
         self.log.debug("Adding start event for build: %s" % build)
-        self.queue_lock.acquire()
+        build.start_time = time.time()
         self.result_event_queue.put(('started', build))
-        self.queue_lock.release()
         self.wake_event.set()
+        self.log.debug("Done adding start event for build: %s" % build)
 
     def onBuildCompleted(self, build):
         self.log.debug("Adding complete event for build: %s" % build)
-        self.queue_lock.acquire()
+        build.end_time = time.time()
+        try:
+            if statsd:
+                key = 'zuul.job.%s' % build.job.name
+                if build.result in ['SUCCESS', 'FAILURE'] and build.start_time:
+                    dt = int((build.end_time - build.start_time) * 1000)
+                    statsd.timing(key, dt)
+                statsd.incr(key)
+        except:
+            self.log.exception("Exception reporting runtime stats")
         self.result_event_queue.put(('completed', build))
-        self.queue_lock.release()
         self.wake_event.set()
+        self.log.debug("Done adding complete event for build: %s" % build)
 
     def reconfigure(self, config):
         self.log.debug("Prepare to reconfigure")
@@ -297,6 +345,7 @@ class Scheduler(threading.Thread):
             self.log.debug("Performing reconfiguration")
             self._init()
             self._parseConfig(self.config.get('zuul', 'layout_config'))
+            self._setupMerger()
             self._pause = False
             self._reconfigure = False
             self.reconfigure_complete_event.set()
@@ -315,6 +364,11 @@ class Scheduler(threading.Thread):
         return False
 
     def run(self):
+        if statsd:
+            self.log.debug("Statsd enabled")
+        else:
+            self.log.debug("Statsd disabled because python statsd "
+                           "package not found")
         while True:
             self.log.debug("Run handler sleeping")
             self.wake_event.wait()
@@ -322,14 +376,14 @@ class Scheduler(threading.Thread):
             if self._stopped:
                 return
             self.log.debug("Run handler awake")
-            self.queue_lock.acquire()
             try:
-                if not self._pause:
-                    if not self.trigger_event_queue.empty():
-                        self.process_event_queue()
-
+                # Give result events priority -- they let us stop builds,
+                # whereas trigger evensts cause us to launch builds.
                 if not self.result_event_queue.empty():
                     self.process_result_queue()
+                elif not self._pause:
+                    if not self.trigger_event_queue.empty():
+                        self.process_event_queue()
 
                 if self._pause and self._areAllBuildsComplete():
                     self._doPauseEvent()
@@ -343,7 +397,6 @@ class Scheduler(threading.Thread):
                         self.wake_event.set()
             except:
                 self.log.exception("Exception in run handler:")
-            self.queue_lock.release()
 
     def process_event_queue(self):
         self.log.debug("Fetching trigger event")
@@ -352,16 +405,29 @@ class Scheduler(threading.Thread):
         project = self.projects.get(event.project_name)
         if not project:
             self.log.warning("Project %s not found" % event.project_name)
+            self.trigger_event_queue.task_done()
             return
 
+        # Preprocessing for ref-update events
+        if hasattr(event, 'refspec'):
+            # Make sure the local git repo is up-to-date with the remote one.
+            # We better have the new ref before enqueuing the changes.
+            # This is done before enqueuing the changes to avoid calling an
+            # update per pipeline accepting the change.
+            self.log.info("Fetching references for %s" % project)
+            self.merger.updateRepo(project)
+
         for pipeline in self.pipelines.values():
+            change = event.getChange(project, self.trigger)
+            if event.type == 'patchset-created':
+                pipeline.manager.removeOldVersionsOfChange(change)
             if not pipeline.manager.eventMatches(event):
                 self.log.debug("Event %s ignored by %s" % (event, pipeline))
                 continue
-            change = event.getChange(project, self.trigger)
             self.log.info("Adding %s, %s to %s" %
                           (project, change, pipeline))
             pipeline.manager.addChange(change)
+        self.trigger_event_queue.task_done()
 
     def process_result_queue(self):
         self.log.debug("Fetching result event")
@@ -370,11 +436,14 @@ class Scheduler(threading.Thread):
         for pipeline in self.pipelines.values():
             if event_type == 'started':
                 if pipeline.manager.onBuildStarted(build):
+                    self.result_event_queue.task_done()
                     return
             elif event_type == 'completed':
                 if pipeline.manager.onBuildCompleted(build):
+                    self.result_event_queue.task_done()
                     return
         self.log.warning("Build %s not found by any queue manager" % (build))
+        self.result_event_queue.task_done()
 
     def formatStatusHTML(self):
         ret = '<html><pre>'
@@ -399,6 +468,34 @@ class Scheduler(threading.Thread):
         ret += '</pre></html>'
         return ret
 
+    def formatStatusJSON(self):
+        data = {}
+        if self._pause:
+            ret = '<p><b>Queue only mode:</b> preparing to '
+            if self._reconfigure:
+                ret += 'reconfigure'
+            if self._exit:
+                ret += 'exit'
+            ret += ', queue length: %s' % self.trigger_event_queue.qsize()
+            ret += '</p>'
+            data['message'] = ret
+
+        data['trigger_event_queue'] = {}
+        data['trigger_event_queue']['length'] = \
+            self.trigger_event_queue.qsize()
+        data['result_event_queue'] = {}
+        data['result_event_queue']['length'] = \
+            self.result_event_queue.qsize()
+
+        pipelines = []
+        data['pipelines'] = pipelines
+        keys = self.pipelines.keys()
+        keys.sort()
+        for key in keys:
+            pipeline = self.pipelines[key]
+            pipelines.append(pipeline.formatStatusJSON())
+        return json.dumps(data)
+
 
 class BasePipelineManager(object):
     log = logging.getLogger("zuul.BasePipelineManager")
@@ -428,6 +525,8 @@ class BasePipelineManager(object):
                 efilters = ''
                 for b in tree.job._branches:
                     efilters += str(b)
+                for f in tree.job._files:
+                    efilters += str(f)
                 if efilters:
                     efilters = ' ' + efilters
                 hold = ''
@@ -501,6 +600,22 @@ class BasePipelineManager(object):
     def enqueueChangesBehind(self, change):
         return True
 
+    def findOldVersionOfChangeAlreadyInQueue(self, change):
+        for c in self.pipeline.getChangesInQueue():
+            if change.isUpdateOf(c):
+                return c
+        return None
+
+    def removeOldVersionsOfChange(self, change):
+        if not self.pipeline.dequeue_on_new_patchset:
+            return
+        old_change = self.findOldVersionOfChangeAlreadyInQueue(change)
+        if old_change:
+            self.log.debug("Change %s is a new version of %s, removing %s" %
+                           (change, old_change, old_change))
+            self.removeChange(old_change)
+            self.launchJobs()
+
     def addChange(self, change):
         self.log.debug("Considering adding change %s" % change)
         if self.isChangeAlreadyInQueue(change):
@@ -513,7 +628,7 @@ class BasePipelineManager(object):
             return False
 
         if not self.enqueueChangesAhead(change):
-            self.log.debug("Failed to enqueue changes ahead of" % change)
+            self.log.debug("Failed to enqueue changes ahead of %s" % change)
             return False
 
         if self.isChangeAlreadyInQueue(change):
@@ -527,6 +642,7 @@ class BasePipelineManager(object):
             if self.start_action:
                 self.reportStart(change)
             change_queue.enqueueChange(change)
+            self.reportStats(change)
             self.enqueueChangesBehind(change)
         else:
             self.log.error("Unable to find change queue for project %s" %
@@ -534,6 +650,37 @@ class BasePipelineManager(object):
             return False
         self.launchJobs()
 
+    def cancelJobs(self, change, prime=True):
+        self.log.debug("Cancel jobs for change %s" % change)
+        to_remove = []
+        for build, build_change in self.building_jobs.items():
+            if build_change == change:
+                self.log.debug("Found build %s for change %s to cancel" %
+                               (build, change))
+                try:
+                    self.sched.launcher.cancel(build)
+                except:
+                    self.log.exception("Exception while canceling build %s "
+                                       "for change %s" % (build, change))
+                to_remove.append(build)
+        for build in to_remove:
+            self.log.debug("Removing build %s from running builds" % build)
+            build.result = 'CANCELED'
+            del self.building_jobs[build]
+
+    def dequeueChange(self, change):
+        self.log.debug("Removing change %s from queue" % change)
+        change_queue = self.pipeline.getQueue(change.project)
+        change_queue.dequeueChange(change)
+
+    def removeChange(self, change):
+        # Remove a change from the queue, probably because it has been
+        # superceded by another change.
+        self.log.debug("Canceling builds behind change: %s because it is "
+                       "being removed." % change)
+        self.cancelJobs(change)
+        self.dequeueChange(change)
+
     def _launchJobs(self, change, jobs):
         self.log.debug("Launching jobs for change %s" % change)
         ref = change.current_build_set.ref
@@ -659,6 +806,7 @@ class BasePipelineManager(object):
                            change)
             change_queue = self.pipeline.getQueue(change.project)
             change_queue.dequeueChange(change)
+            self.reportStats(change)
             return True
 
     def reportChange(self, change):
@@ -692,9 +840,9 @@ class BasePipelineManager(object):
     def formatReport(self, changeish):
         ret = ''
         if self.pipeline.didAllJobsSucceed(changeish):
-            ret += 'Build successful\n\n'
+            ret += self.pipeline.success_message + '\n\n'
         else:
-            ret += 'Build failed\n\n'
+            ret += self.pipeline.failure_message + '\n\n'
 
         if changeish.dequeued_needing_change:
             ret += "This change depends on a change that failed to merge."
@@ -704,16 +852,23 @@ class BasePipelineManager(object):
                    "rebase your change and upload a new patchset."
         else:
             if self.sched.config.has_option('zuul', 'url_pattern'):
-                pattern = self.sched.config.get('zuul', 'url_pattern')
+                url_pattern = self.sched.config.get('zuul', 'url_pattern')
             else:
-                pattern = None
+                url_pattern = None
             for job in self.pipeline.getJobs(changeish):
                 build = changeish.current_build_set.getBuild(job.name)
                 result = build.result
-                if result == 'SUCCESS' and job.success_message:
-                    result = job.success_message
-                elif result == 'FAILURE' and job.failure_message:
-                    result = job.failure_message
+                pattern = url_pattern
+                if result == 'SUCCESS':
+                    if job.success_message:
+                        result = job.success_message
+                    if job.success_pattern:
+                        pattern = job.success_pattern
+                elif result == 'FAILURE':
+                    if job.failure_message:
+                        result = job.failure_message
+                    if job.failure_pattern:
+                        pattern = job.failure_pattern
                 url = None
                 if build.url:
                     if pattern:
@@ -832,6 +987,37 @@ class BasePipelineManager(object):
         ret = ret.format(**locals())
         return ret
 
+    def reportStats(self, change):
+        if not statsd:
+            return
+        try:
+            # Update the guage on enqueue and dequeue, but timers only
+            # when dequeing.
+            if change.dequeue_time:
+                dt = int((change.dequeue_time - change.enqueue_time) * 1000)
+            else:
+                dt = None
+            changes = len(self.pipeline.getAllChanges())
+
+            # stats.timers.zuul.pipeline.NAME.resident_time
+            # stats_counts.zuul.pipeline.NAME.total_changes
+            # stats.gauges.zuul.pipeline.NAME.current_changes
+            key = 'zuul.pipeline.%s' % self.pipeline.name
+            statsd.gauge(key + '.current_changes', changes)
+            if dt:
+                statsd.timing(key + '.resident_time', dt)
+                statsd.incr(key + '.total_changes')
+
+            # stats.timers.zuul.pipeline.NAME.ORG.PROJECT.resident_time
+            # stats_counts.zuul.pipeline.NAME.ORG.PROJECT.total_changes
+            project_name = change.project.name.replace('/', '.')
+            key += '.%s' % project_name
+            if dt:
+                statsd.timing(key + '.resident_time', dt)
+                statsd.incr(key + '.total_changes')
+        except:
+            self.log.exception("Exception reporting pipeline stats")
+
 
 class IndependentPipelineManager(BasePipelineManager):
     log = logging.getLogger("zuul.IndependentPipelineManager")
@@ -959,6 +1145,11 @@ class DependentPipelineManager(BasePipelineManager):
                                                            changes))
         return changes
 
+    def _unableToMerge(self, change, all_changes):
+        self.log.info("Unable to merge changes %s" % all_changes)
+        self.pipeline.setUnableToMerge(change)
+        self.possiblyReportChange(change)
+
     def _launchJobs(self, change, jobs):
         self.log.debug("Launching jobs for change %s" % change)
         ref = change.current_build_set.ref
@@ -968,13 +1159,15 @@ class DependentPipelineManager(BasePipelineManager):
             dependent_changes = self._getDependentChanges(change)
             dependent_changes.reverse()
             all_changes = dependent_changes + [change]
-            commit = self.sched.merger.mergeChanges(all_changes, ref)
-            if not commit:
-                self.log.info("Unable to merge changes %s" % all_changes)
-                self.pipeline.setUnableToMerge(change)
-                self.possiblyReportChange(change)
+            if (dependent_changes and
+                not dependent_changes[-1].current_build_set.commit):
+                self._unableToMerge(change, all_changes)
                 return
+            commit = self.sched.merger.mergeChanges(all_changes, ref)
             change.current_build_set.commit = commit
+            if not commit:
+                self._unableToMerge(change, all_changes)
+                return
         #TODO: remove this line after GERRIT_CHANGES is gone
         dependent_changes = self._getDependentChanges(change)
         for job in jobs:
@@ -1015,6 +1208,15 @@ class DependentPipelineManager(BasePipelineManager):
                            (change.change_behind, change))
             self.cancelJobs(change.change_behind, prime=prime)
 
+    def removeChange(self, change):
+        # Remove a change from the queue (even the middle), probably
+        # because it has been superceded by another change (or
+        # otherwise will not merge).
+        self.log.debug("Canceling builds behind change: %s because it is "
+                       "being removed." % change)
+        self.cancelJobs(change)
+        self.dequeueChange(change, keep_severed_heads=False)
+
     def handleFailedChange(self, change):
         # A build failed. All changes behind this change will need to
         # be retested. To free up resources cancel the builds behind
@@ -1034,13 +1236,13 @@ class DependentPipelineManager(BasePipelineManager):
                            "failure." % change)
             self.cancelJobs(change_behind, prime=False)
 
-    def dequeueChange(self, change):
+    def dequeueChange(self, change, keep_severed_heads=True):
         self.log.debug("Removing change %s from queue" % change)
         change_ahead = change.change_ahead
         change_behind = change.change_behind
         change_queue = self.pipeline.getQueue(change.project)
         change_queue.dequeueChange(change)
-        if not change_ahead and not change.reported:
+        if keep_severed_heads and not change_ahead and not change.reported:
             self.log.debug("Adding %s as a severed head" % change)
             change_queue.addSeveredHead(change)
         self.dequeueDependentChanges(change_behind)
